use std::slice::Iter;

use crate::error::ParserError;
use crate::token::{Token, TokenType};

#[derive(Debug, PartialEq)]
pub struct Select {
    pub fields: Vec<Field>,
    pub source: Identifier,
}

#[derive(Debug, PartialEq)]
pub struct Identifier {
    pub name: String,
}

#[derive(Debug, PartialEq)]
pub enum Field {
    Star,
}

impl Identifier {
    pub fn new(name: &str) -> Self {
        Self {
            name: name.to_string(),
        }
    }
}

pub fn parse(tokens: Vec<Token>) -> Result<Select, ParserError> {
    select_expression(&mut tokens.iter())
}

fn select_expression(tokens: &mut Iter<Token>) -> Result<Select, ParserError> {
    eat(tokens, TokenType::Select)?;
    let fields = eat(tokens, TokenType::Star).and_then(|_| Ok(vec![Field::Star]))?;
    eat(tokens, TokenType::From)?;
    let source = eat(tokens, TokenType::Identfiier).and_then(|t| Ok(Identifier::new(&t.value)))?;

    Ok(Select { fields, source })
}

fn eat<'a>(tokens: &'a mut Iter<Token>, token_type: TokenType) -> Result<&'a Token, ParserError> {
    match tokens.next() {
        Some(token) => {
            if token.token_type == token_type {
                Ok(token)
            } else {
                Err(ParserError::new(&format!(
                    "Expected {} but got {}",
                    token_type, token.token_type
                )))
            }
        }
        None => Err(ParserError::new("Unexpected end of tokens")),
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    use crate::lexer::scan;

    fn _parse(input: &str) -> Result<Select, ParserError> {
        let tokens = scan(input).unwrap();
        parse(tokens)
    }

    #[test]
    fn it_parses_a_simple_select() {
        assert_eq!(
            _parse("SELECT * FROM index").unwrap(),
            Select {
                fields: vec![Field::Star],
                source: Identifier::new("index")
            },
        )
    }
}
